// Compile in MinGW with the following command:
// g++ -o mygame.dll template.cpp -shared -Wl,--enable-auto-import -Wl,--enable-runtime-pseudo-reloc -s -lalleg
// ...or something like that...

#include <IGame.h>
#include <cmath>
#include <list>
//#include <MASkinG.h>
using namespace std;


/*
	Game design
		>You will be a blue sphere in a sea filled with red spheres of Doom.
		>The red spheres will float around without friction.
		>If you're hit by a red sphere you will loose radius.
		>Red-Red and Red-Blue collisions result in changed direction for both spheres.
		>You can shoot black spheres of redemption at the red spheres.
		>When a red sphere is hit by a black sphere it will shrink and change direction.
		>When a red sphere has shrunk to a rather small size, it will change into a green sphere.
		>Green spheres can be picked up to regain radius.
		?>If green collides with red, red gains radius.
		>You move around with direction keys.
		>You aim and fire with mouse.

		TODO:
		 Reflective collision bounce. (Maybe not so important, they bounce quite nice for the gameplay anyway.)
		 Start level setting. (Maybe I need something more from the amgc interface to do this.)

	Changes
		1.3
			Added return true to the init function.
			Difficulty curve adjusted.
			Added settings.
		1.2
			Increasing difficulty and an indicator for level of difficulty.
			Score now depends on level.
			Changed the DLLEXPORT thing.
			Draw green/red circle within the blue to indicate max/min radius.
*/

#define MIN_RADIUS 8
#define MAX_RADIUS 16

class Sphere{
public:
	enum Spheretype{BLUE, RED, BLACK, GREEN};

	static int s_w, s_h, score, level;
	static int colour[4];

	Sphere(Spheretype t, float ix, float iy, float idx, float idy, float iradius){
		type=t;
		x=ix;
		y=iy;
		dx=idx;
		dy=idy;
		radius=iradius;
		doomed=0;
	}
	//Return collision truth.
	bool Collide(Sphere &other){
		float dx=other.x-x;
		float dy=other.y-y;
		float tr=radius+other.radius;
		float force=tr*tr-(dx*dx+dy*dy);
		if(force>0){
			float angle=atan2(dy, dx);
			HandleCollision(other.type, angle, force);
			other.HandleCollision(type, angle+M_PI, force);
			return true;
		}
		return false;
	}
	//Return 0 if ok, 1 for removal.
	void HandleCollision(Spheretype t, float angle, float force){
		switch(type){
		case BLUE:
			if(t==RED){
				Sphere::score-=10*level;
				radius--;
				if(radius<MIN_RADIUS){
					doomed=true;
					return;
				}
				//Back out first;
				x-=dx;
				y-=dy;
				//Change direction	(temp formula, the bounce should be reflective)
				dx=-cos(angle);//*force;
				dy=-sin(angle);//*force;
			}
			else if(t==GREEN){
				if(radius<MAX_RADIUS)
					radius++;
				else
					; //Maybe reward player with more powerfull shots, to use with rmb.
			}
			break;
		case RED:
			if(t==BLUE){
				radius--;
				if(radius<MIN_RADIUS)
					type=GREEN;
			}
			if(t==RED){
				radius--;
				if(radius<MIN_RADIUS)
					type=GREEN;
			}
			if(t==BLACK){
				Sphere::score+=level+1;
				radius--;
				if(radius<MIN_RADIUS)
					type=GREEN;
			}
			if(t==BLUE || t==RED || t==BLACK){
				//Back out first;
				x-=dx;
				y-=dy;
				//Change direction	(temp formula, the bounce should be reflective)
				dx=-cos(angle);//*force;
				dy=-sin(angle);//*force;

/*				angle+=M_PI;
				if(angle>=M_PI*2)
					angle-=M_PI*2;

				float a=atan2(dy, dx)+M_PI;
				if(a>=M_PI*2)
					a-=M_PI*2;
				float s=sqrt(dx*dx+dy*dy);
				float d1 = ((a - angle) + M_PI*4);
				while(d1>=M_PI*2)
					d1-=M_PI*2;
				d1-=M_PI;
				angle+=2*-d1;

				dx=cos(angle)*s;
				dy=sin(angle)*s;
				x+=dx;
				y+=dy;
*/			}
			if(t==GREEN){
				if(radius<MAX_RADIUS)
					radius++;
			}
			break;
		case BLACK:
			if(t==RED){
				doomed=true;
			}
			break;
		case GREEN:
			if(t==BLUE || t==RED){
				doomed=true;
			}
			break;
		}
	}
	bool Doomed(){
		return doomed;
	}
	void Draw(BITMAP *canvas){
		circle(canvas, int(x), int(y), int(radius), colour[type]);
		if(type==BLUE){
			if(radius>MAX_RADIUS-1)
				circle(canvas, int(x), int(y), int(radius/2), colour[GREEN]);
			if(radius<MIN_RADIUS+1)
				circle(canvas, int(x), int(y), int(radius/2), colour[RED]);
		}
		//Mirroring code commented since I don't use wrapping
/*		if(x<radius)
			circle(canvas, int(x+s_w), int(y), int(radius), colour[type]);
		if(x+radius>s_w)
			circle(canvas, int(x-s_w), int(y), int(radius), colour[type]);
		if(y<radius)
			circle(canvas, int(x), int(y+s_h), int(radius), colour[type]);
		if(y+radius>s_h)
			circle(canvas, int(x), int(y-s_h), int(radius), colour[type]);
*/	}
	void Update(){
		x+=dx;
		y+=dy;
		if((x<radius && dx<0) || (x+radius>s_w && dx>0)){
			dx=-dx;
			x+=dx;
			if(type==BLACK)
				radius>1?--radius:doomed=true;
		}
		if((y<radius && dy<0) || (y+radius>s_h && dy>0)){
			dy=-dy;
			y+=dy;
			if(type==BLACK)
				radius>1?--radius:doomed=true;
		}
		//Wrapping code commented, I didn't feel like programming collision across wrapped edges.
/*		if(x<0)
			x+=s_w;
		if(x>=s_w)
			x-=s_w;
		if(y<0)
			y+=s_h;
		if(y>=s_h)
			y-=s_h;
*/	}
	float x, y, dx, dy, radius;
private:
	Spheretype type;
	//Doomed indicates that the sphere is doomed to be removed.
	bool doomed;
};

int Sphere::s_w=0;
int Sphere::s_h=0;
int Sphere::score=0;
int Sphere::level=0;
int Sphere::colour[4];



GameSettings sod_settings[] = {
	{ "startlevel", "Level at start", "0|1|2|3|4|5|6|7|8|9", 0 },
	{ "orientation", "Control", "Screen|Mouse", 0 },
	{ 0, 0, 0, 0 }
};

/***********************
_.-'                '-._
	Spheres of Doom.
'-._                _.-'
**********************/
class SoD : public IGame {
	public:
		SoD() : IGame() {
			player=NULL;
			uT=0;
		}

		~SoD() {
		}

		char *GetName() {
			return "Spheres of Doom";
		}

		char *GetAuthor() {
			return "Anders Andersson";
		}

		char *GetDescription() {
			return "This is my game... of Doom! Move around with keys and use the mouse to aim and fire. Left mouse button can be held or clicked for regular fire. Right mouse button sacrifises a radius to fire a powerfull blast. You are blue, they are red. Pick up the green if you're loosing radius.";
		}

		char *GetVersion() {
			return "1.3";
		}

		char *GetIconPath() {
			return "sod.bmp";
		}

		bool Init(){//HWND wnd, int bpp, int mode, int s_w, int s_h) {
			if (!IGame::Init()) {
//			if (!IGame::Init(wnd, bpp, mode, s_w, s_h)) {
				return false;
			}


			//Hardcoded icon...
/*			BITMAP *bmp=create_bitmap(32, 32);
			PALETTE pal;
			get_palette(pal);
			if(bmp){
				clear_to_color(bmp, makecol(255, 255, 255));
				circle(bmp, 16, 16, 15, makecol(0, 0, 255));
				circle(bmp, 16, 16, 12, makecol(255, 0, 0));
				circle(bmp, 16, 16, 9, makecol(0, 255, 0));
				circle(bmp, 16, 16, 4, makecol(0, 0, 0));
				save_bmp("sod.bmp", bmp, pal);
				destroy_bitmap(bmp);
			}
*/
			// custom initialization goes here
			// ...
			uT=0;
//			Sphere::s_w=s_w;
//			Sphere::s_h=s_h;
			Sphere::s_w=SCREEN_W;
			Sphere::s_h=SCREEN_H;
			Sphere::score=0;
			Sphere::level=0;
			Sphere::colour[Sphere::BLUE]=makecol(0, 0, 255);
			Sphere::colour[Sphere::RED]=makecol(255, 0, 0);
			Sphere::colour[Sphere::BLACK]=makecol(0, 0, 0);
			Sphere::colour[Sphere::GREEN]=makecol(0, 255, 0);

			Sphere::level=sod_settings[0].value;
			spawnlimit=1500-Sphere::level*100;
			levelT=0;
			spawnT=0;
			shootT=0;

			mp=0;
			player=NULL;
			spheres.push_front(Sphere(Sphere::BLUE, Sphere::s_w/2, Sphere::s_h/2, 0, 0, MAX_RADIUS));
			player=&(*(spheres.begin()));
			return true;
		}

		void Deinit() {
			// custom deinitialization goes here
			// ...

			// no more Allegro calls after this line!
			IGame::Deinit();
		}

		void Draw(BITMAP *canvas) {
			// Do all the drawing here. Draw only to the canvas bitmap.
			// ...
			clear_to_color(canvas, makecol(255, 255, 255));
			for(list<Sphere>::iterator i=spheres.begin(); i!=spheres.end(); ++i){
				i->Draw(canvas);
			}

			textprintf_ex(canvas, font, 16, 8, makecol(0,0,0), -1, "Score: %i", Sphere::score);
			textprintf_right_ex(canvas, font, Sphere::s_w - 16, 8, makecol(0,0,0), -1, "Level: %i", Sphere::level);
/*			if(!player){
				textout_centre_ex(canvas, font, "You have died!", Sphere::s_w/2, Sphere::s_h/2, makecol(0,0,0), -1);
				textout_centre_ex(canvas, font, "press esc to quit or click mouse for new game", Sphere::s_w/2, Sphere::s_h/2+16, makecol(0,0,0), -1);
			}
*/		}

		int Logic(int keys[8], int mouse[4], float dT) {
			// Do all input processing and game logic here. Refer to the
			// comments in IGame.h for details on the format of the input
			// parameters and the requirements for the output.
			// ...

/*			if(!player && (mp && !(mouse[2]&1))){
				spheres.clear();
				spheres.push_front(Sphere(Sphere::BLUE, Sphere::s_w/2, Sphere::s_h/2, 0, 0, MAX_RADIUS));
				player=&(*(spheres.begin()));
				Sphere::score=0;
			}
			mp=mouse[2]&1;
*/
			if(player && player->radius>MIN_RADIUS && (mp && !(mouse[2]&2))){
				--player->radius;
				for(float angle=0; angle<2*M_PI; angle+=M_PI/8)
					spheres.push_back(Sphere(Sphere::BLACK, player->x, player->y, cos(angle)*4, sin(angle)*4, MIN_RADIUS/4*3));
			}
			mp=mouse[2]&2;

			//Time dependant updating
			uT+=dT;
			spawnT+=dT;
			shootT+=dT;
			levelT+=dT;
			if(uT<1)
				return 0;
			uT-=1;

			if(mouse[2]&1 && shootT>=200){
				shootT-=200;
				if(player){
					float angle=atan2(mouse[1]-player->y, mouse[0]-player->x);
					spheres.push_back(Sphere(Sphere::BLACK, player->x, player->y, cos(angle)*8, sin(angle)*8, MIN_RADIUS/2));
				}
			}
			if(!(mouse[2]&1)){
				shootT=200;
			}


			if(levelT>500){
				levelT-=500;
				spawnlimit--;
			}
			//Redspawn
			if(spawnT>spawnlimit){
				spawnT-=spawnlimit;
				if(spawnlimit>500){
					Sphere::level=9-int(spawnlimit-500)/100;
				}
				float angle=float(rand()%360)*M_PI/180;
				int side=rand()%4;
				int pos=rand()%Sphere::s_w;
				spheres.push_back(Sphere(Sphere::RED,
				 side<2?pos:side==2?-MAX_RADIUS:Sphere::s_w+MAX_RADIUS,
				 side>1?pos:side=1?-MAX_RADIUS:Sphere::s_h+MAX_RADIUS,
				 cos(angle), sin(angle), MAX_RADIUS));
			}

			if(player){
				if(sod_settings[1].value==0){ //Screen oriented
					if(keys[0])
						player->dx-=.05;
					if(keys[1])
						player->dx+=.05;
					if(keys[2])
						player->dy-=.05;
					if(keys[3])
						player->dy+=.05;
				}
				else{
					float angle=atan2(mouse[1]-player->y, mouse[0]-player->x);
					if(keys[0]){
						player->dx-=cos(angle+M_PI/2)*.05;
						player->dy-=sin(angle+M_PI/2)*.05;
					}
					if(keys[1]){
						player->dx+=cos(angle+M_PI/2)*.05;
						player->dy+=sin(angle+M_PI/2)*.05;
					}
					if(keys[2]){
						player->dx+=cos(angle)*.05;
						player->dy+=sin(angle)*.05;
					}
					if(keys[3]){
						player->dx-=cos(angle)*.05;
						player->dy-=sin(angle)*.05;
					}
				}
			}

			for(list<Sphere>::iterator i=spheres.begin(); i!=spheres.end(); ++i){
				i->Update();
			}
			for(list<Sphere>::iterator i=spheres.begin(); i!=spheres.end(); ){
				for(list<Sphere>::iterator j=i; j!=spheres.end(); ++j){
					if(j==i)
						continue;
					i->Collide(*j);
				}
				bool irem=false;
				for(list<Sphere>::iterator j=i; j!=spheres.end(); ){
					if(j->Doomed()){
						if(i==j)
							irem=true;
						if(&(*j)==player)
							player=NULL;
						j=spheres.erase(j);
					}
					else
						++j;
				}
				if(irem)
					i=spheres.begin();
				else
					++i;
			}
			if(!player)
				return 1;
			return 0;
		}

		int Score() {
			// Return -1 if this game doesn't have points, otherwise the last
			// valid score where more points is better.
			return Sphere::score;
		}

		bool UseSystemMouse() {
			// Return false if the game draws its own cursor or doesn't use
			// the mouse at all.
			return true;
		}

		void Pause() {
		}

		void Resume() {
		}

		// Should return true if keys can be redefined (otherwise
		// global keys are used)
		bool RedefinesKeys() {
			return true;
		}

		// Should return names of redefinable keys or NULL if
		// key can't be redefined. i = [0,7]
		const char *KeyName(int i) {
			switch (i) {
				case 0:
					return "Left";
					break;
				case 1:
					return "right";
					break;
				case 2:
					return "up";
					break;
				case 3:
					return "down";
					break;
				default: return 0;
			};
		}

		// Should return to a NULL terminated array of GameSettings structures.
		// For example you could define the array like this:
		//
		//	GameSettings my_settings[] = {
		//		{ "name1", "option 1", "first|second|third", 1 },
		//		{ "name2", "option 2", "first|second|third", 0 },
		//		{ "name3", "option 3", "first|second|third", 2 },
		//		{ 0, 0, 0, 0 }
		//	};
		//
		// and then return my_settings in this function.
		GameSettings* GetSettings() {
			return sod_settings;
		}
	private:
		bool mp;
		float uT, spawnT, levelT, shootT;
		float spawnlimit;
		Sphere *player;
		list<Sphere> spheres;

};



extern "C" DLLEXPORT IGame* GetPlugin() {
	return new SoD;
}
